#include <gtk/gtk.h>
#include <libical/ical.h>
#include "ical-util.h"
#include "koto-task.h"
#include "koto-date-combo.h"
#include "koto-priority-combo.h"
#include "koto-task-editor.h"

G_DEFINE_TYPE (KotoTaskEditor, koto_task_editor, GTK_TYPE_TABLE);

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), KOTO_TYPE_TASK_EDITOR, KotoTaskEditorPrivate))

typedef struct
{
  KotoTask *task;
  gboolean dirty;
  /* TODO: not sure if I really need this */
  gboolean ignore;
} KotoTaskEditorPrivate;

enum {
  PROP_0,
  PROP_TASK,
  PROP_DIRTY,
};

#define FIELD_DATA "KotoTaskEditor:Field"
#define PROP_DATA "KotoTaskEditor:Property"
#define WIDGET_DATA "KotoTaskEditor:Widget"

/* Forward declaration for function pointer typedef */
typedef struct _DisplayField DisplayField;

/* Function that creates a widget for a particular field */
typedef GtkWidget* (*EditorNew) (KotoTaskEditor *editor);
/* Function that sets the widget value for a particular field/property */
typedef void (*EditorSetFunc) (GtkWidget *editor, icalproperty *prop);

/*
 * Struct defining a field to display.
 */
struct _DisplayField {
  icalproperty_kind kind;
  char *display_name;
  EditorNew editor_new;
  EditorSetFunc set_value;
};

static GtkWidget *entry_new (KotoTaskEditor *editor);
static void entry_set_from_string (GtkWidget *widget, icalproperty *prop);
static GtkWidget *text_new (KotoTaskEditor *editor);
static void text_set_from_string (GtkWidget *widget, icalproperty *prop);
static GtkWidget *priority_new (KotoTaskEditor *editor);
static void priority_set (GtkWidget *widget, icalproperty *prop);
static GtkWidget *date_new (KotoTaskEditor *editor);
static void date_set (GtkWidget *widget, icalproperty *prop);

/*
 * List of fields to display in the main editor.  The order of this list is the
 * order they will be shown in.
 */
static const DisplayField fields[] = {
  { ICAL_SUMMARY_PROPERTY, "_Summary:", entry_new, entry_set_from_string },
#if 0
  /* TODO: replace with combo */
  { ICAL_CATEGORIES_PROPERTY, "_Category:", entry_new, entry_set_from_string },
#endif
  { ICAL_PRIORITY_PROPERTY, "_Priority:", priority_new, priority_set },
  { ICAL_DUE_PROPERTY, "_Due Date:", date_new, date_set },
  /* TODO: add button to open URL */
  { ICAL_URL_PROPERTY, "_Web Address:", entry_new, entry_set_from_string },
  { ICAL_DESCRIPTION_PROPERTY, "_Description:", text_new, text_set_from_string },
};

/*
 * Callback when a GtkEntry field changes.
 */
static void
on_entry_changed (GtkEntry *entry, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  DisplayField *field;
  icalproperty *prop;
  const char *text;

  if (priv->ignore)
    return;

  field = g_object_get_data (G_OBJECT (entry), FIELD_DATA);
  prop = g_object_get_data (G_OBJECT (entry), PROP_DATA);

  if (!field) {
    g_warning (FIELD_DATA " property not set");
    return;
  }

  if (!prop) {
    prop = icalproperty_new (field->kind);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (entry), PROP_DATA, prop);
  }

  text = gtk_entry_get_text (entry);
  /* NO means use the default value type. Ugly as sin. */
  icalproperty_set_value_from_string (prop, text, "NO");

  priv->dirty = TRUE;
  return;
}

/*
 * Editor creation function (an EditorNew) , which creates a plain GtkEntry.
 */
static GtkWidget *
entry_new (KotoTaskEditor *editor)
{
  GtkWidget *entry;
  entry = gtk_entry_new ();
  g_signal_connect (entry, "changed", G_CALLBACK (on_entry_changed), editor);
  return entry;
}

/*
 * Editor setter function (an EditorSetFunc), that sets the value of a GtkEntry
 * to the string value of @prop.
 */
static void
entry_set_from_string (GtkWidget *widget, icalproperty *prop)
{
  const char *text = NULL;
  if (prop) {
    text = icalproperty_get_value_as_string (prop);
  }
  
  gtk_entry_set_text (GTK_ENTRY (widget), text ?: "");

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

static void
on_text_changed (GtkTextBuffer *buffer, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  GtkWidget *widget;
  GtkTextIter start, end;
  DisplayField *field;
  icalproperty *prop;
  const char *text;

  if (priv->ignore)
    return;

  widget = g_object_get_data (G_OBJECT (buffer), WIDGET_DATA);
  if (!widget) {
    g_warning (WIDGET_DATA " property not set");
    return;
  }

  field = g_object_get_data (G_OBJECT (widget), FIELD_DATA);
  prop = g_object_get_data (G_OBJECT (widget), PROP_DATA);

  if (!field) {
    g_warning (FIELD_DATA " property not set");
    return;
  }
  
  if (!prop) {
    prop = icalproperty_new (field->kind);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
  }

  gtk_text_buffer_get_start_iter (buffer, &start);
  gtk_text_buffer_get_end_iter (buffer, &end);

  text = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);

  /* NO means use the default value type. Ugly as sin. */
  icalproperty_set_value_from_string (prop, text, "NO");
  
  priv->dirty = TRUE;
  return;
}

static GtkWidget *
text_new (KotoTaskEditor *editor)
{
  GtkWidget *scrolled, *textview;
  GtkTextBuffer *buffer;

  scrolled = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled), GTK_SHADOW_IN);

  textview = gtk_text_view_new ();
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview), GTK_WRAP_WORD);
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  g_object_set_data (G_OBJECT (buffer), WIDGET_DATA, scrolled);
  gtk_widget_show (textview);
  gtk_container_add (GTK_CONTAINER (scrolled), textview);

  g_signal_connect (gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview)),
                    "changed", G_CALLBACK (on_text_changed),
                    editor);
  
  return scrolled;
}

static void
text_set_from_string (GtkWidget *widget, icalproperty *prop)
{
  const char *text = NULL;
  char *s;
  GtkWidget *textview;
  GtkTextBuffer *buffer;

  if (prop) {
    text = icalproperty_get_value_as_string (prop);
  }

  textview = gtk_bin_get_child (GTK_BIN (widget));
  if (!GTK_IS_TEXT_VIEW (textview)) {
    g_warning ("Child widget was not GtkTextView");
    return;
  }
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  if (text) {
    s = g_strcompress (text);
    gtk_text_buffer_set_text (buffer, s, -1);
    g_free (s);
  } else {
    gtk_text_buffer_set_text (buffer, "", -1);
  }
  
  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

static void
on_priority_changed (KotoPriorityCombo *combo, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  icalproperty *prop;
  int priority;

  if (priv->ignore)
    return;

  prop = g_object_get_data (G_OBJECT (combo), PROP_DATA);
  
  if (!prop) {
    prop = icalproperty_new (ICAL_PRIORITY_PROPERTY);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, prop);
  }

  priority = koto_priority_combo_get_priority (combo);
  if (priority != PRIORITY_NONE) {
    icalproperty_set_priority (prop, priority);
  } else {
    icalcomponent_remove_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, NULL);
  }

  priv->dirty = TRUE;
}

static GtkWidget *
priority_new (KotoTaskEditor *editor)
{
  GtkWidget *widget;
  widget = koto_priority_combo_new ();
  g_signal_connect (widget, "changed", G_CALLBACK (on_priority_changed), editor);
  return widget;
}

static void
priority_set (GtkWidget *widget, icalproperty *prop)
{
  int priority = 0;
  if (prop) {
    priority = icalproperty_get_priority (prop);
  }
  
  koto_priority_combo_set_priority (KOTO_PRIORITY_COMBO (widget), priority);

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

static void
on_date_notify (KotoDateCombo *combo, GParamSpec *spec, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  icalproperty *prop;
  GDate *date;
  icaltimetype due;

  if (priv->ignore)
    return;

  prop = g_object_get_data (G_OBJECT (combo), PROP_DATA);
  
  if (!prop) {
    prop = icalproperty_new (ICAL_DUE_PROPERTY);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, prop);
  }

  date = koto_date_combo_get_date (combo);
  if (g_date_valid (date)) {
    /* This may be the fastest way from GDate to icaltimetype */
    due = icaltime_from_day_of_year (g_date_get_day_of_year (date),
                                     g_date_get_year (date));
    icalproperty_set_due (prop, due);
  } else {
    icalcomponent_remove_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, NULL);
  }

  priv->dirty = TRUE;
}

static GtkWidget *
date_new (KotoTaskEditor *editor)
{
  GtkWidget *widget;
  widget = koto_date_combo_new ();
  g_signal_connect (widget, "notify::date", G_CALLBACK (on_date_notify), editor);
  return widget;
}

static void
date_set (GtkWidget *widget, icalproperty *prop)
{
  GDate date;
  struct icaltimetype due;

  g_date_clear (&date, 1);
  
  if (prop) {
    due = icalproperty_get_due (prop);
    g_date_set_dmy (&date, due.day, due.month, due.year);
  }
  
  koto_date_combo_set_date (KOTO_DATE_COMBO (widget), &date);

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

/*
 * Called for every widget when a new icalcomponent is set.  If this widget has
 * an associated DisplayField set, then updates it with the value of the
 * relevant property.
 */
static void
update_field (GtkWidget *widget, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  DisplayField *field;
  icalproperty *prop;

  field = g_object_get_data (G_OBJECT (widget), FIELD_DATA);
  if (!field)
    return;

  prop = icalcomponent_get_first_property (priv->task->comp, field->kind);
  
  if (field->set_value) {
    field->set_value (widget, prop);
  } else {
    g_warning ("Field data for %s doesn't define set_value", field->display_name);
  }
}


/*
 * GObject methods.
 */

static void
koto_task_editor_get_property (GObject *object, guint property_id,
                              GValue *value, GParamSpec *pspec)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_TASK:
    g_value_set_boxed (value, priv->task);
    break;
  case PROP_DIRTY:
    g_value_set_boolean (value, priv->dirty);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
koto_task_editor_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_TASK:
    priv->task = g_value_dup_boxed (value);
    priv->dirty = FALSE;
    /* Now iterate over every widget, updating any fields with values from this icalcomponent. */
    priv->ignore = TRUE;
    gtk_container_foreach (GTK_CONTAINER (object), (GtkCallback)update_field, object);
    priv->ignore = FALSE;
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
koto_task_editor_dispose (GObject *object)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);
  
  if (priv->task) {
    koto_task_unref (priv->task);
    priv->task = NULL;
  }

  if (G_OBJECT_CLASS (koto_task_editor_parent_class)->dispose)
    G_OBJECT_CLASS (koto_task_editor_parent_class)->dispose (object);
}

static void
koto_task_editor_finalize (GObject *object)
{
  G_OBJECT_CLASS (koto_task_editor_parent_class)->finalize (object);
}

static void
koto_task_editor_class_init (KotoTaskEditorClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (KotoTaskEditorPrivate));

  object_class->get_property = koto_task_editor_get_property;
  object_class->set_property = koto_task_editor_set_property;
  object_class->dispose = koto_task_editor_dispose;
  object_class->finalize = koto_task_editor_finalize;

  g_object_class_install_property (object_class, PROP_TASK,
                                   g_param_spec_boxed ("task", "task", NULL,
                                                       KOTO_TYPE_TASK,
                                                       G_PARAM_READWRITE |
                                                       G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));
  g_object_class_install_property (object_class, PROP_DIRTY,
                                   g_param_spec_boolean ("dirty", "dirty", NULL,
                                                         FALSE,
                                                         G_PARAM_READABLE |
                                                         G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));
}

static void
koto_task_editor_init (KotoTaskEditor *self)
{
  int i;

  g_object_set (self,
                "homogeneous", FALSE,
                "row-spacing", 6,
                "column-spacing", 6,
                NULL);
  
  for (i = 0; i < G_N_ELEMENTS (fields); i++) {
    const DisplayField *field;
    GtkWidget *label, *widget;
    
    field = &fields[i];
    
    label = gtk_label_new_with_mnemonic (field->display_name);
    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
    gtk_widget_show (label);
    gtk_table_attach (GTK_TABLE (self), label, 0, 1, i+1, i+2, GTK_FILL, 0, 0, 0);

    widget = field->editor_new (self);
    g_object_set_data (G_OBJECT (widget), FIELD_DATA, (gpointer)field);
    gtk_widget_show (widget);
    gtk_table_attach_defaults (GTK_TABLE (self), widget, 1, 2, i+1, i+2);
    gtk_label_set_mnemonic_widget (GTK_LABEL (label), widget);
  }
}


/*
 * Public methods.
 */

GtkWidget *
koto_task_editor_new (void)
{
  return g_object_new (KOTO_TYPE_TASK_EDITOR, NULL);
}

gboolean
koto_task_editor_is_dirty (KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv;

  g_return_val_if_fail (KOTO_IS_TASK_EDITOR (editor), FALSE);

  priv = GET_PRIVATE (editor);

  return priv->dirty;
}
